##############################################################################
# Copyright 2008, Gerhard Weis
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#  * Neither the name of the authors nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT
##############################################################################
'''
this module implements all de/serialisers defied in WSDL 1 for mime encoding.
'''
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.image import MIMEImage
from email.generator import Generator
from email import message_from_file
from StringIO import StringIO

from lxml import etree
from zope.interface import classProvides

from rsl.globalregistry import lookupimpl
from rsl.interfaces import ISerializer, IDeserializer
from rsl.xsd.interfaces import IXMLDeserializer, IXMLSerializer

def splitmime(mimetype):
    '''
    split a mime type name into two parts.
    e.g.: text/xml => (text, xml)
    '''
    if mimetype is None:
        return None, None
    mtsplit = mimetype.split('/')
    if len(mtsplit) == 1:
        if mtsplit[0] in ('', '*'):
            return None, None # no major defined
        else:
            return mtsplit[0], None
    if mtsplit[0] in ('', '*'):
        if mtsplit[1] in ('', '*'):
            return None, None
        else:
            return None, mtsplit[1]
    if mtsplit[1] in ('', '*'):
        return mtsplit[0], None
    return mtsplit[0], mtsplit[1]

# some predefined majore mime types and their classes.
mimeclsmap = {'text': MIMEText,
              'application': MIMEApplication,
              'audio': MIMEAudio,
              'image': MIMEImage}
        
def getmimeclass(mimetype):
    '''
    helper function to resolve class in mimclsmap for mimetype.
    '''
    major, minor = splitmime(mimetype)
    cls = mimeclsmap.get(major.lower(), None)
    return major, minor, cls
     
# 'mime:content'
class MimeContentSerializer(object):
    '''
    implement mime:content en/decoding rules.
    '''
    
    classProvides(ISerializer, IDeserializer)
    
    @classmethod
    def serialize(cls, params, operationinfo, **kw):
        '''
        params    ... dict of values to serialise
        operationinfo ... IOperationInfo instance
        mimeobj   ... if this is set and true, then return created MIMEBase instance
        '''
        # only one part allowd here...
        # part['type'] ... mime-type
        # part['mimetype'] ... mime-type
        # part['name'] ... name of part in params dict
        # only one part allowed here or no part for empty messages
        if len(operationinfo.input) > 0:
            pinfo = operationinfo.input[0]
        else: 
            # if we don't serialize anything, then we have no content-type?
            return {}, '' 
        value = params[pinfo.name]
        if pinfo.type:
            value = operationinfo.typesystem.getType(pinfo.type).encode(value) # what about complex types? should not happen here
        if pinfo.serializer == 'application/x-www-form-urlencoded':
            msg = MIMENonMultipart('application', 'x-www-form-urlencoded')
            serialiser = lookupimpl(ISerializer, 'urlEncoded')
            _, value = serialiser.serialize(params, operationinfo)
            msg.set_payload(value)
        else:
            major, minor, cls = getmimeclass(pinfo.serializer)
            if cls is None:
                msg = MIMENonMultipart(major, minor)
                msg.set_payload(value)
            else:
                msg = cls(value, minor)
        if kw.get('mimeobj', False):
            return msg
        outstr = StringIO()
        mimegenerator = Generator(outstr)
        mimegenerator._dispatch(msg) # not that nice to use private methods, but there is no separate content/header method
        return dict(msg), outstr.getvalue()
    
    @classmethod
    def deserialize(cls, data, operationinfo, **kw):
        '''
        extracts parameter names and values from mime-encoded data.
        '''
        # this must be some text/xml or whatever namespace.
        # 1. check for mime-type:
        #     which one to use?, the one in kw['headers'] or the one from partsinfo
        # only one part allowed here
        pinfo = operationinfo.output[0]
        deserializer = lookupimpl(IDeserializer, pinfo.serializer)
        if deserializer is not None:
            return deserializer.deserialize(data, operationinfo, **kw) 
        else:
            # ok... nothing to do, return what we have
            return kw['headers'], data
    
# 'mime:mimeXml'
class MimeXmlSerializer(object):
    '''
    implement mime:mimeXml en/decoding rules.
    '''
    
    classProvides(ISerializer, IDeserializer)
    
    @classmethod
    def serialize(cls, params, operationinfo, **kw):
        '''
        params    ... dict of values to serialise
        operationinfo ... IOperationInfo instance
        mimeobj   ... if this is set and true, then return created MIMEBase instance
        '''
        # only one part allowed here
        pinfo = operationinfo.input[0]
        value = params[pinfo.name]
        serializer = lookupimpl(IXMLSerializer, pinfo.serializer)
        _, encvalue = serializer.serialize(value, pinfo.type, operationinfo.typesystem, None)
        msg = MIMEText(encvalue, 'xml')
        if kw.get('mimeobj', False):
            return msg
        outstr = StringIO()
        mimegenerator = Generator(outstr)
        mimegenerator._dispatch(msg) # not that nice to use private methods, but there is no separate content/header method
        return dict(msg), outstr.getvalue()
    
    @classmethod
    def deserialize(cls, data, operationinfo, **_):
        '''
        extracts parameter names and values from xml-encoded data.
        '''
        # this must be some text/xml for whatever namespace.
        # 1. parse xml:
        datatree = etree.fromstring(data)
        # only one part allowed here?
        pinfo = operationinfo.output[0]
        serialiser = lookupimpl(IXMLDeserializer, pinfo.serializer)
        value = serialiser.deserialize(datatree, pinfo.type, operationinfo.typesystem)
        return None, value
            
    
# 'mime:multipartRelated'
class MimeMultipartRelatedSerializer(object):
    '''
    implement mime:multipartRelated en/decoding rules.
    '''
    
    classProvides(ISerializer, IDeserializer)
    
    @classmethod
    def serialize(cls, params, operationinfo, **kw):
        '''
        params    ... dict of values to serialise
        operationinfo ... IOperationInfo
        mimeobj   ... if this is set and true, then return created MIMEBase instance
        '''
        # this is quite some overhaed but very generic :)
        # serialise each part....
        # attach all parts to mimemultipart...
        # serialise this again and return it.
        msg = MIMEMultipart('related')
        for pinfo in operationinfo.input:
            serialiser = lookupimpl((), ISerializer, pinfo.serializer)
            partval = serialiser.serialize(params, operationinfo, mimeobj=True)
            if isinstance(partval, MIMENonMultipart):
                msg.attach(partval)
            else:
                # TODO: for now assume it is a text/xml
                mnp = MIMENonMultipart(None, None)
                del mnp['Content-Type']
                if partval[0] is not None:
                    for key, value in partval[0]:
                        mnp.add_header(key, value)
                mnp.set_payload(partval[1])                    
                msg.attach(mnp)
        if kw.get('mimeobj', False):
            return msg
        outstr = StringIO()
        mimegenerator = Generator(outstr)
        mimegenerator._dispatch(msg) # not that nice to use private methods, but there is no separate content/header method
        return dict(msg), outstr.getvalue()
    
    @classmethod
    def deserialize(cls, data, operationinfo, **kw):
        '''
        extracts parameter names and values from mime-multipart-encoded data.
        '''
        # 1. rebuild multipart message
        headers = kw['headers']
        buf = StringIO()
        for k,v in headers.items():
            buf.write('%s: %s\r\n' % (k,v))
        buf.write('\r\n')
        buf.write(data)
        buf.seek(0)
        mpm = message_from_file(buf)
        value = []
        i = 0 # index into multipart message
        for pinfo in operationinfo.output:
            serialiser = lookupimpl((), IDeserializer, pinfo.serializer)
            partmsg = mpm.get_payload()[i]
            partheaders, partvalue = serialiser.deserialize(partmsg.get_payload(), operationinfo, headers=dict(partmsg))
            value.append((partheaders,partvalue))
        return None, partvalue



##############################################################################
# Copyright 2008, Gerhard Weis
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#  * Neither the name of the authors nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT
##############################################################################
'''
This module defines the WSDL mime extension elements.
'''
from zope.interface import directlyProvides

from rsl.wsdl1.interfaces import IWSDLExtensionFactory

from rsl.globalregistry import lookupimpl
from rsl.implementations import ParamInfo
from rsl.wsdl1.binding import WSDLParamInfo
from rsl.wsdl1.interfaces import IParamInfo
from rsl.misc.namespace import clark2tuple

# mime:content  part?,type?
# mime:mimeXml part?
# mime:multipartRelated
#    mime:part
#       soap:body
#       mime:content
#       mime:mimeXml
class MIMEParamInfo(WSDLParamInfo):
    '''
    extend WSDL specific I/O parameter infos with mime specific information.
    '''
    
    def __init__(self, operation):
        '''
        initialise empty instance.
        '''
        super(MIMEParamInfo, self).__init__(operation)
        self.exttag = None
        self.type = None
        self.part = None
        self.subparts = None
    
    def frometree(self, inputelem): #also used for outputelem
        '''
        parse information about I/O parameters from etree instance. 
        ''' 
        super(MIMEParamInfo, self).frometree(inputelem)
        _, self.exttag = clark2tuple(inputelem[0].tag)
        # check whether tag is mimeContent, mimeXml or multipartRelated
        if self.exttag == 'content':
            # TODO: there can be more than one content tags
            self.part = inputelem[0].get('part')
            self.type = inputelem[0].get('type')
            self.getMsgPart() # test if part exists.
        elif self.exttag == 'mimeXml':
            self.part = inputelem[0].get('part')
            self.type = 'text/xml'
        elif self.exttag == 'multipartRelated':
            self.type = 'multipart/related'
            for partelem in inputelem[0]:
                # parse mime:part elements
                childns, _ = clark2tuple(partelem[0])
                pif = lookupimpl(IParamInfo, childns)
                subpif = pif(self.operation).frometree(partelem) # maybe it's not nice to pass something different than wsdl:input/output/fault to this factory
                if self.subparts is None:
                    self.subparts = []
                self.subparts.append(subpif)
        return self
    
    def getSerializer(self):
        '''
        return name of serializer for WSDL-mime
        '''
        if self.exttag == 'content' and self.type=='application/x-www-form-urlencoded':
            return self.type
        return 'mime:' + self.exttag
    
    def getMsgPart(self):
        '''
        helper method to find all related and valid msgParts.
        '''
        # for http binding there can only be urlEncoded or urlReplacement and all parts for a
        # message are used.
        porttype = self.operation.binding.wsdl.porttypes[self.operation.binding.porttype]
        porttypeoperation = porttype[self.operation.name]
        message = self.operation.binding.wsdl.messages[porttypeoperation[self.tag]['message']]
        if self.part is not None:
            for p in message:
                if p['name'] == self.part:
                    return p
        else:
            if len(message) > 1:
                #warn('Illegal mime binding. specify part attribute or make sure there is only one part in the message.')
                raise ValueError('Illegal mime binding. specify part attribute or make sure there is only one part in the message.')
            elif len(message) > 0:
                return message[0] # assume there is only one part
            else:
                return None # empty message is possible
        raise ValueError('Invalid WSDL file. Can not find msg. part %s.' % self.part)

    def getParamInfo(self):
        '''
        return IParamInfo instance filled with information from this instance.
        '''
        pinfos = []
        if self.exttag == 'multipartRelated':
            for subpart in self.subparts:
                pinfos.extend(subpart.getParamInfo())
        else:
            msgpart = self.getMsgPart()
            if msgpart is not None:
                pinfo = ParamInfo()
                pinfo.name = msgpart['name']
                if self.type == 'text/xml':
                    pinfo.serializer = 'xml'
                    if 'type' in msgpart:
                        pinfo.type = msgpart['type']
                        pinfo.serializer = pinfo.serializer + ':type'
                    elif 'element' in msgpart:
                        pinfo.type = msgpart['element']
                        pinfo.serializer = pinfo.serializer + ':element'
                else:
                    pinfo.type = msgpart['type']
                pinfos.append(pinfo)
        return pinfos

def mimeparaminfofactory(operation, inputelem):
    '''
    Factory for MIMEParamInfo instances.
    '''
    paraminfo = MIMEParamInfo(operation)
    paraminfo = paraminfo.frometree(inputelem)
    return paraminfo
directlyProvides(mimeparaminfofactory, IWSDLExtensionFactory)



#!/usr/bin/env python
##############################################################################
# Copyright 2008-2009, Gerhard Weis
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#  * Neither the name of the authors nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT
##############################################################################
import os
from setuptools import setup
from setuptools import find_packages
# for setuptools see: http://peak.telecommunity.com/DevCenter/setuptools

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(name = 'rsl.mime',
      version = '0.2.3',
      package_dir={'': 'src'},
      packages = find_packages('src', exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
      include_package_data = True,
      namespace_packages = ['rsl'],
      
      # dependencies:
      install_requires = ['rsl >= 0.2.1',
                          'rsl.xsd >= 0.2.1',
                          'rsl.wsdl1 >= 0.2.2',
                          'zope.interface',
                          'lxml >= 2.0',
                          'setuptools'],
      
      # PyPI metadata
      author='Gerhard Weis',
      author_email='gerhard.weis@proclos.com',
      description='Remote Service Library mime encoding module',
      license = 'BSD',
      #keywords = '',
      url='http://rslib.sourceforge.net',

      long_description=read('README.txt') +
                       read('CHANGES.txt'),
      
      #download_url='',
      classifiers = ['Development Status :: 3 - Alpha',
                     'Environment :: Web Environment',
                     'Intended Audience :: Developers',
                     'License :: OSI Approved :: BSD License',
                     'Operating System :: OS Independent',
                     'Programming Language :: Python',
                     'Topic :: Internet',
                     'Topic :: Software Development :: Libraries :: Python Modules',
                     'Topic :: System :: Networking',
                     ],
      entry_points = {'rsl.register': ['method = rsl.mime.factories:register'],
                      'rsl.wsdl1.binding.operation.input' : [ 'http://schemas.xmlsoap.org/wsdl/mime/ = rsl.mime.wsdl1:mimeparaminfofactory' ],
                      'rsl.wsdl1.binding.operation.output' : [ 'http://schemas.xmlsoap.org/wsdl/mime/ = rsl.mime.wsdl1:mimeparaminfofactory' ] }
     )
